#include "crogue.h"

//{{{
static int FontHeight(uint c)
{
#ifdef IS_CALCULATOR
	return 6;
#else
	return 1;
#endif
}
//}}}
//{{{
static int CharWidth(char c)
{
#ifdef IS_CALCULATOR
	return FontCharWidth(c);
#else
	return 1;
#endif
}
//}}}
//{{{
static int StringWidth(char *str)
{
	uint len=0;
	
	while(*str != '\0') {
		len += CharWidth(*str);
		str++;
	}
	
	return len;
}
//}}}

void clear_string(int x, int y, int width, int height)
{
#ifdef IS_CALCULATOR
	SCR_RECT rect = (SCR_RECT){{x, y, x+width-1, y-1+height*FontHeight(F_4x6)}};
	Black_ScrRectFill(&rect, &rect, A_REVERSE);
#else
	int i, j;
	for(i=x; i<x+width; i++)
		for(j=y; j<y+height; j++)
		{
			gotoxy(i, j);
			addch(' ');
		}
#endif
}

static int numTabStops = 0;
static const uchar *tabStops = NULL;

void setTabStops(int num, const uchar *value)
{
	numTabStops = num;
	tabStops = value;
}

static int findTab(int pos)
{
	int i;
	for(i=0; i<numTabStops; i++)
	{
		if(tabStops[i] > pos)
			return tabStops[i];
	}
	return SCREEN_COLUMNS * 2;
}

/*
 * Rows=0 means no wrapping allowed; instead, clip the right edge (don't pause, either)
 * Tab stops are given as distance from x, not from the left edge of the screen
 * Rows are in textual lines, width, x, and y are in pixels.
 */
void draw_string(const char *string, draw_string_info *start_state, short x, short y, short width, short rows, short wrap)
{
	int i, ii;
	int curX=0;
	int tempX;
	int curRow=0;
	int maxWidth;
	
	if(start_state)
	{
		curX = start_state->curX;
		curRow = start_state->curRow;
	}
	
	tempX = curX;
	
	i=0;
	
	while( string[i] != '\0' )
	{
		maxWidth = width;
		if( wrap && curRow+1 == rows )
			maxWidth -= StringWidth(" -More-");
		
		if( string[i] == '\t' )
		{
			if(findTab(tempX) > maxWidth)
				goto startNewLine;
			else
				tempX = findTab(tempX);
		}
		else
		{
			tempX += CharWidth(string[i]);
		}
		
		// If past the end of the line
		if(tempX >= maxWidth || string[i] == '\n') {
startNewLine:
			if(!wrap)
				break;
			
			if(string[i] != '\n' && string[i] != '\t')
			{
				// Back up until (a) it fits and (b) it isn't in the middle of a word
				for( ; i >= 0 && (string[i] != ' ' || width>maxWidth); i-- )
					tempX -= CharWidth(string[i]);
			}
			
			// Print it out
			for(ii=0; ii<i; ii++) {
				if(string[ii]!='\t')
					Black_DrawChar(x+curX, y + curRow*FontHeight(F_4x6), string[ii], A_NORMAL);
				if(string[ii]=='\t')
					curX = findTab(curX);
				else
					curX += CharWidth(string[ii]);
			}
			
			// If out of lines, print the -More- prompt
			if( curRow+1 == rows ) {
				Black_DrawStr(curX+x, curRow*FontHeight(F_4x6) + y, " -More-", A_NORMAL);
				READCHAR;
				clear_string(x, y, width, rows);
				
				curRow = -1;
				curX = tempX = 0;
			}
			
			string += i;
			string++;
			
			i=0;
			
			curRow++;
			curX = tempX = 0;
		} else {
			i++;
		}
	}
	
	// Print the remainder
	for( i=0; string[i]!='\0'; i++ ) {
		if(string[i] == '\t')
		{
			curX = findTab(curX);
			continue;
		}
		Black_DrawChar( x+curX, y + curRow*FontHeight(F_4x6), string[i], A_NORMAL);
		curX += CharWidth( string[i] );
	}
	
	if(start_state)
	{
		start_state->curX = curX;
		start_state->curRow = curRow;
	}
}
