#ifndef TIROGUE_H
#define TIROGUE_H


#include "config.h"

#include "machdep.h"

//{{{
// Data dereferencing
#define CONSTFILEDEREF(a)   (deref_file_ptr(a))
#define isNull(a)           ((a).offset == 0)

#define MDESC(a)            find_monster_description(a)
#define MATTACK(a, b)       (MDESC(a)->attacks[b])

#define TILEDESC(a)         (w->tiledescs[(a).type])

#define ITEMDESCN(a)        (*((itemdesc*)CONSTFILEDEREF( w->itemdescs[a] )))
#define ITEMDESC(a)         (ITEMDESCN((a).type))
#define INVENTORY(a)        (w->plr.inventory[(a)])
//}}}

#include "rle.h"
#include "util.h"
#include "constdata.h"
#include "map.h"
#include "items.h"
#include "monst.h"
#include "ui.h"
#include "keyboard.h"
#include "player.h"
#include "automatic.h"	// Header automatically generated from data


//{{{
typedef struct tile	// 6 bytes
{
	uchar type;
	uchar flags;
	uchar special;
	uchar numitems;
	uint topitem;
} tile;

#define TFLAG_EXPLORED        0x01
#define TFLAG_HIDETRAP        0x02
#define TFLAG_OCCUPIED        0x04
#define TFLAG_ROOM            0x08
#define TFLAG_LIT             0x10
#define TFLAG_INTRINSIC_LIGHT 0x20
//}}}

#define MONSTERS_MAX 128

//{{{
typedef struct world
{	
	world_descriptor desc;
	
#ifdef IS_CALCULATOR
	HANDLE constfile;		// TIOS-provided handle to the constant data file
#endif // IS_CALCULATOR
	char* constfileoffset;	//   Pointer to the start of the constant data file. Add this to a file offset (constant data types use offsets) to get a pointer to the data at
							// that offset.
	tiledesc*  tiledescs;	//   Pointer to table of tile descriptors which specifies what the tiles look like and qualities such as whether they are transparent to vision, etc.
	mapdesc*   mapdescs;	//   Pointer to table of map descriptors which specify how a map should be generated, what maps it has stairs leading too, and other properties.
	filelink*  itemdescs;	//   Pointer to table of item descriptors which specify the properties of an items given a type number (usually from a struct item).
	spelldesc* spelldescs;
	classdesc* playerclasses;
	filelink*  shuffledata;
	uint*      shuffletranslation;
	
	tile t[MAPSIZE_Y][MAPSIZE_X];
	
	itemsinfo items;
	
	monster m[MONSTERS_MAX];
	player plr;
	
	ulong *itemids;			// One BIT per item type, indicating whether or not
							// it's been identified.
	
	uchar level;		// The level number (index in mapdescs) the player is
						// currently on.
	uchar maxlevel;		// The highest level number the player has ever been on
	
	
	ulong time;			// Number of player-turns into the game
	
	UI_Tilefield tilefield;			//   State info for widget which displays the tiled main area of the screen
	uchar messagevis;	// Whether there should be a message displayed in the message field (used to decide when to clear old messages).
} world;
//}}}


// Game data
#ifdef REGISTERVAR
	register world *w asm("a4");
#else
	world *w;
#endif

// Constant data
#ifdef IS_CALCULATOR
extern const uchar tiles[128*5];    // Kept in code because it's not the same
                                    // on different calculators
extern const uchar tilemask[5];
extern const uchar blacktile[5];
extern const uchar whitetile[5];
#else
extern const uchar tiles[128];		// For PC, this is mapping from calc-side tiles to extended ASCII
#endif

extern const schar dir[8][2];	// 8 cardinal directions converted into (x,y) form

// Functions

//{{{
// Draw.c
void calc_light();
void draw();		// Redraw the screen
void message(const char *format, ...);
void draw_all_tiles(void);
void draw_tile(ushort x, ushort y);

// Files.c
void save_game(void);
int restore_game(void);
void cleanup_save(void);
int load_level(int source, coord pos);
void save_level(void);

// Magic.c
int bolt(int x, int y, direction dir, int distance);
void throw_item_from_menu(void);
void throw_readied_item(void);
void use_item(void);
void magic_map(void);
void teleport_player(void);
void identify(void);

// Tirogue.c
void fatal(char *msg);
void quit(void) ATTRIB(noreturn);
void die(int m);
void timepass(int amt);
void satiate(int amt);
direction prompt_dir(char *msg);
											// are with constant parameters and can be evaluated at compile-time
// Prompt.c
int prompt(char *msg);
int prompt_letter(char *msg);
direction prompt_dir(char *msg);

// Debug.c
void debug(void);

// Drawstring.c
typedef struct draw_string_info
{
	int curX;
	int curRow;
} draw_string_info;

void setTabStops(int num, const uchar *value);
void clear_string(int x, int y, int width, int height);
void draw_string(const char *string, draw_string_info *start_state, short x, short y, short width, short rows, short wrap);

// Spell.c
void read_spellbook(int i);
void init_spells(void);
void cast_spell(void);
//}}}

#endif
