#include "crogue.h"

/*
Parameters:
pos: top-left corner of the window
size: distance from top-left corner to bottom-left corner
numchoices: number of entries on the menu
getchoice: pointer to function taking int N and returning Nth entry on the list

Return codes:
-2 Error
-1 User aborted (no selection)
0 to numchoices: user selection #
*/

#ifdef IS_CALCULATOR

void UI_Menu_Redraw( coord pos, coord size, int numchoices, void (*getchoice)(int N, char *buf), int scrollpos, int rows );

//{{{
int UI_Menu_Pick( coord pos, coord size, int numchoices, void (*getchoice)(int N, char *buf), int viewonly )
{
	int rows;
	int in;
	int selected=0, offset=0;
	int redraw;
	SCR_RECT rect;
	
	rows = size.y/6;
	
	UI_Menu_Redraw(pos, size, numchoices, getchoice, 0, rows);
	
	if( !viewonly ) {
		rect = (SCR_RECT){{pos.x+1, pos.y+selected*6+1, pos.x+size.x-1, pos.y+selected*6+6}};
		Black_ScrRectFill(&rect, &rect, A_XOR);
	}
	
	while(1) {
		redraw = 0;
		
		in = READCHAR;
		
		if( !viewonly ) {
			rect = (SCR_RECT){{pos.x+1, pos.y+(selected-offset)*6+1, pos.x+size.x-1, pos.y+(selected-offset)*6+6}};
			Black_ScrRectFill(&rect, &rect, A_XOR);
		}
		
		if(in==KEY_UP)
		{
			if( viewonly ) {
				redraw=1;
				offset--;
				if(offset<0) offset=0;
			} else {
				if(selected==0)
					selected=numchoices-1;
				else
					selected--;
			}
		} else if(in==KEY_DOWN)
		{
			if( viewonly ) {
				redraw=1;
				offset++;
				if(offset+rows >= numchoices) offset=numchoices-rows;
			} else {
				selected++;
				if(selected>=numchoices)
					selected=0;
			}
		} else if(in==KEY_UP+4096)	// 2nd+Up
		{
			selected -= rows;
			if( selected<0 )
				selected = 0;
			offset -= rows;
			if( offset<0 )
				offset = 0;
			redraw = 1;
		} else if(in==KEY_DOWN+4096)	// 2nd+Down
		{
			selected += rows;
			if( selected>=numchoices )
				selected = numchoices-1;
			offset += rows;
			if( offset+rows>=numchoices )
				offset = numchoices-rows;
			redraw = 1;
		} else if(in==KEY_CONFIRM) {
			return selected;
		} else if(in==KEY_ESC) {
			return -1;
		}
		
		if(!viewonly)
		{
			if(selected>=offset+rows) {
				offset += rows/2;
				if(offset+rows>=numchoices) offset = numchoices-rows;
				redraw=1;
			}
			if(selected<=offset) {
				offset -= rows/2;
				if(offset<0) offset=0;
				redraw=1;
			}
			
			if( selected-offset < 0 || selected-offset >= rows ) {
				offset = selected - rows/2;
				if( offset<0 ) offset=0;
				if( offset+rows > numchoices ) offset = numchoices - rows;
			}
		}
		if(redraw) {
			UI_Menu_Redraw(pos, size, numchoices, getchoice, offset, rows);
		}
		
		if( !viewonly ) {
			rect = (SCR_RECT){{pos.x+1, pos.y+(selected-offset)*6+1, pos.x+size.x-1, pos.y+(selected-offset)*6+6}};
			Black_ScrRectFill(&rect, &rect, A_XOR);
		}
	}
	
	return 0;
}
//}}}
//{{{
void UI_Menu_Redraw( coord pos, coord size, int numchoices, void (*getchoice)(int N, char *buf), int scrollpos, int rows )
{
	char buf[256];
	int i;
	SCR_RECT rect = {{pos.x, pos.y, pos.x+size.x, pos.y+size.y}};
	
	SetCurClip(&rect);
	
	// Clear a space
	Black_ScrRectFill( &rect, &rect, A_REVERSE );
	
	// Draw a frame
	Black_DrawLine(pos.x, pos.y, pos.x+size.x, pos.y, A_NORMAL);
	Black_DrawLine(pos.x, pos.y, pos.x, pos.y+size.y, A_NORMAL);
	Black_DrawLine(pos.x+size.x, pos.y, pos.x+size.x, pos.y+size.y, A_NORMAL);
	Black_DrawLine(pos.x, pos.y+size.y, pos.x+size.x, pos.y+size.y, A_NORMAL);
	
	// Draw contents
	for(i=scrollpos; i<numchoices && i<rows+scrollpos; i++) {
		getchoice(i, buf);
		draw_string(buf, NULL, pos.x+2, pos.y+6*(i-scrollpos)+2, size.x, 1, 0);
	}
}
//}}}
#else
//{{{
void UI_Menu_Redraw( int numchoices, void (*getchoice)(int N, char *buf), int scrollpos )
{
	int i;
	char outbuf[128];
	char buf[256];
	
	for(i=scrollpos; i<min(scrollpos+18, numchoices); i++)
	{
		getchoice(i, buf);
		sprintf(outbuf, "%c - %s", i-scrollpos+'a', buf);
		clear_line(i+2);
		draw_string(outbuf, NULL, 0, i+2, SCREEN_COLUMNS, 1, 0);
	}
	clear_line(i+2);
}
//}}}

//{{{
int UI_Menu_Pick( coord pos, coord size, int numchoices, void (*getchoice)(int N, char *buf), int viewonly )
{
	int scrollpos=0;
	int in;
	
	while(1)
	{
		UI_Menu_Redraw(numchoices, getchoice, scrollpos);
		
		in = READCHAR;
		
		if(in==' ' || in=='>') {
			scrollpos += 17;
			if(scrollpos>numchoices)
				return -1;
		}
		else if(in=='<') {
			scrollpos -= 17;
			if(scrollpos<0) scrollpos=0;
		}
		else if(islower(in)) {
			if(scrollpos+in-'a' < 0 || scrollpos+in-'a' >= numchoices)
				continue;
			clrtiles();
			return scrollpos+in-'a';
		}
	}
}
//}}}

#endif
