#include "crogue.h"

/*
//{{{
void init_map(coord size)
{
	int i;
	
	w->mapsize = size;
	
	w->t = (tile**)malloc(sizeof(tile*) * size.y);
	if( !w->t )
	fatal("Out of memory.");
	
	for(i=0; i<size.y; i++)
	{
		w->t[i] = (tile*)malloc(sizeof(tile) * size.x);
		if( !w->t[i] )
			fatal("Out of memory.");
	}
}
//}}}
//{{{
void reinit_map(coord size)
{
	cleanup_map();
	init_map(size);
}
//}}}
//{{{
void cleanup_map(void)
{
	int i;
	
	if( !w->t ) return;
	
	for(i=0; i<w->mapsize.y; i++) {
		if( w->t[i] )
			free( w->t[i] );
		else
			break;
	}
	
	free( w->t );
}
//}}}
*/

static const schar simpledir[4][2] = {
			{0,-1}, {1,0}, {0,1}, {-1,0}	// Ordered clockwise
	};
static const schar cornerdir[4][2] = {
		{-1,-1},		{1,-1},
		
		{-1,1},			{1,1}
	};
const schar dir[8][2] = {
		{-1,-1},	{0,-1},		{1,-1},
		{-1,0},					{1,0},
		{-1,1},		{0,1},		{1,1}
	};

//{{{
typedef struct room
{
	int left, right, top, bottom;
} room;
//}}}

void place_trap(int x, int y);
void set_rect(int top, int bot, int left, int right, int v);
void replace(int old, int n);
void make_dungeon(dungeondesc *desc);
void make_cavern(cavedesc *desc);
void make_fixedmap(pregenmapdesc *desc);
void fill_pass(void);
void connect_pass(coord pos);
int connectareas(int num_areas);
void fill_room(room r);

//{{{
void gen_map(int source, coord pos)
{
	int x, y;
	map_gen_method *m;
	
	m = (map_gen_method*)CONSTFILEDEREF(w->mapdescs[w->level].genmethod);
	
	cleanup_items();
	init_items();
	
	// Clear tiles
	for(y=0; y<MAPSIZE_Y; y++)
		for(x=0; x<MAPSIZE_X; x++) {
			w->t[y][x].type = 0;
			w->t[y][x].flags = 0;
			w->t[y][x].numitems = 0;
			w->t[y][x].topitem = 0;
		}
	
	// Clear monster table
	for(y=0; y<MONSTERS_MAX; y++) {
		w->m[y].type = filelink_null;
		w->m[y].energy = 0;
	}
	
	switch(m->type)
	{
		case map_dungeon:
			make_dungeon(&m->ext.dungeon);
			break;
		case map_cave:
			make_cavern(&m->ext.cave);
			break;
		case map_pregen:
			make_fixedmap(&m->ext.pregen);
			break;
	}
	
	connect_pass(pos);
	
	place_player_on_map(source, pos);
	fill_pass();
	
	w->plr.score += (w->mapdescs[w->level].difficulty - 1) * 50;
}
//}}}
//{{{
void place_trap(int x, int y)
{
	if(w->level < 2) return; // No traps on first level
	w->t[y][x].type = TILE_DARTTRAP;
	w->t[y][x].flags |= TFLAG_HIDETRAP;
}
//}}}
//{{{
void make_fixedmap(pregenmapdesc* desc)
{
	int xi, yi;
	
	rle_set_source(rle_pointer, desc->tiles);
	
	for(yi=0; yi<MAPSIZE_Y; yi++)
		for(xi=0; xi<MAPSIZE_X; xi++)
		{
//			w->t[yi][xi].type = desc->tiles[yi][xi];
			w->t[yi][xi].type = rlegetc();
			
			if(desc->globallight)
				w->t[yi][xi].flags |= TFLAG_INTRINSIC_LIGHT;
		}
}
//}}}
//{{{
void fill_pass(void)
{
	int yi=0, xi=0;
	
	// Place items and traps
	for(; yi<MAPSIZE_Y; yi++)
		for(xi=0; xi<MAPSIZE_X; xi++)
		{
			if(w->t[yi][xi].type==TILE_FLOOR)
			{
				// 1 in 100 chance for a trap per tile
				if(rand()%100 == 0) {
					place_trap(xi, yi);
				}
			}
			if(w->tiledescs[w->t[yi][xi].type].passable)
			{
				// One in 25 chance for an item on each tile
				if(xrandom(0x1900)==0) {
					// Place an item
					place_item(randitem(), xi, yi);
				}
				
				// One in 40 chance for a monster on each tile
				if(xrandom(0x2800)==0) {
					// Place a monster
					addmonster(xi, yi);
				}
			}
		}
}
//}}}
//{{{
void connect_pass(coord pos)
{
	int x, y, i;
	spec_descriptor *specs;
	
	specs = (spec_descriptor*)CONSTFILEDEREF(w->mapdescs[ w->level ].special);
	
	// Place connections
	for( i=0; i < w->mapdescs[ w->level ].numspecial; i++ )
	{
		if( specs[i].location.x==0 && specs[i].location.y == 0 ) {	// Random location
			while(1) {
				x=xrandom(((MAPSIZE_X-2)<<8)+2);
				y=xrandom(((MAPSIZE_Y-2)<<8)+2);
				if( w->t[y][x].type==TILE_FLOOR )
					break;
			}
		} else {
			x = specs[i].location.x;
			y = specs[i].location.y;
		}
		w->t[y][x].type = specs[i].type;
		w->t[y][x].special = i;
	}
}
//}}}
//{{{
void place_player_on_map(int source, coord pos)
{
	int x, y;
	spec_descriptor *connections;
	
	connections = ((spec_descriptor*)CONSTFILEDEREF(w->mapdescs[w->level].special));
	
	// Place the player
	if( pos.x==0 && pos.y==0 )
	{
		for(x=0; x<MAPSIZE_X; x++)
			for(y=0; y<MAPSIZE_Y; y++) {
				if( w->tiledescs[w->t[y][x].type].is_connection && connections[w->t[y][x].special].destmap == source )
				{
					w->plr.x = x;
					w->plr.y = y;
					w->t[y][x].flags |= TFLAG_OCCUPIED;
					return;
				}
			}
	}
	else
	{
		w->plr.x = pos.x;
		w->plr.y = pos.y;
	}
	
	place_player_randomly();
}
//}}}
//{{{
void make_dungeon(dungeondesc *desc)
{
	int num_areas;
	int i, x, y;
	int left, right, top, bot;
	int type, otype, result1, result2;
	int retrycount;
	
	room rooms[32];

start:
	num_areas = desc->num_rooms;
	retrycount = 0;
	
	// Clear tiles
	for(y=0; y<MAPSIZE_Y; y++)
		for(x=0; x<MAPSIZE_X; x++)
			w->t[y][x].type = 0;
	
	// Place rooms randomly
	for(i=0; i<num_areas; i++) {
retry:
		left = RANGE(0x3A, 1);
		right = min(left+RANGE(8,3), MAPSIZE_X-2);
		top = RANGE(16, 1);
		bot = min(top+RANGE(6,3), MAPSIZE_Y-2);
		
		// Check for collisions with existing rooms
		for(y=top-1; y<=bot+1; y++)
			for(x=left-1; x<=right+1; x++)
				if( w->t[y][x].type != 0 )
				{
					retrycount++;
					if( retrycount>1000 )
						goto start;
					else
						goto retry;
				}
		
		for(y=top; y<=bot; y++)
			for(x=left; x<=right; x++)
				w->t[y][x].type=i+1;
		
		rooms[i].left   = left;
		rooms[i].right  = right;
		rooms[i].top    = top;
		rooms[i].bottom = bot;
	}
	
	// Place 'holes'
	num_areas+=desc->num_holes;
	for(i=0; i<desc->num_holes; i++)
		w->t[ RANGE(MAPSIZE_Y-2, 1) ][ RANGE(MAPSIZE_X-2, 1) ].type = desc->num_rooms+1+i;
	
	// Connect areas
	if( !connectareas(num_areas) ) {
		goto start;
	}
	
	// Convert into tiles
	for(y=0; y<MAPSIZE_Y; y++)
		for(x=0; x<MAPSIZE_X; x++)
		{
			if( w->t[y][x].type )
				w->t[y][x].type = TILE_FLOOR;
			else
				w->t[y][x].type = TILE_WALL;
		}
	
	// Make doors (where applicable)
	// A tile becomes a door if...
	// No doors on adjacent tiles
	// Two opposing EDGE tiles are walls
	// Two opposing EDGE tiles are empty space
	// At least one CORNER tile is space
	// At least one CORNER tile is a wall
	for(y=1; y<MAPSIZE_Y-1; y++)
		for(x=1; x<MAPSIZE_X-1; x++)
		{
			// Simplified version of first three rules above: simpledir rotation clockwise alternates between wall and non-wall
			otype =  w->t[y+simpledir[0][1]][x+simpledir[0][0]].type;
			if(otype==TILE_CDOOR) goto notadoor;
			
			for(i=1; i<4; i++) {
				type = w->t[y+simpledir[i][1]][x+simpledir[i][0]].type;
				if(type == otype || type == TILE_CDOOR || type==TILE_SDOOR || type==TILE_LDOOR) {
					goto notadoor;
				}
				otype = type;
			}
			
			result1=result2=0;
			for(i=0; i<3; i++)
			{
				if( w->t[y+cornerdir[i][1]][x+cornerdir[i][0]].type==TILE_FLOOR )
					result1=1;
				if( w->t[y+cornerdir[i][1]][x+cornerdir[i][0]].type==TILE_WALL )
					result2=1;
			}
			
			if(result1 && result2) {
				if(RANGE(5,0)==0)	// 1 in 5 chance a door becomes secret
					w->t[y][x].type = TILE_SDOOR;
				else
					w->t[y][x].type = TILE_CDOOR;
			}
notadoor:
			;
		}
	
	for(i=0; i<desc->num_rooms; i++)
		fill_room(rooms[i]);
}
//}}}
//{{{
void make_cavern(cavedesc *desc)
{
	int i, ii, iii, x, y, state, adjcount, areacount;

	// Initialize randomly
	for(y=1; y<MAPSIZE_Y-1; y++)
		for(x=1; x<MAPSIZE_X-1; x++)
			w->t[y][x].type = ( xrandom(0x6400)>=45 ) ? 1 : 0;

	// Apply modified Game of Life rules
	// Hack (as usual): 'type' field used for 'real' state, 'flags' field used for 'new' state
	for(i=0; i<7; i++) // 10 generations
	{
		for(y=1; y<MAPSIZE_Y-1; y++)
			for(x=1; x<MAPSIZE_X-1; x++)
			{
				adjcount=0;
				for(ii=-1; ii<=1; ii++)
					for(iii=-1; iii<=1; iii++)
						if( !w->t[y+ii][x+iii].type )
							adjcount++;
				
				w->t[y][x].flags = !(adjcount>=5);
			}

		for(y=1; y<MAPSIZE_Y-1; y++)
			for(x=1; x<MAPSIZE_X-1; x++)
				w->t[y][x].type = w->t[y][x].flags;
	}
	
	// FIXME: Either the area location, room-merging, both, or their interaction is broken.
	
	// Locate areas
	// Each row gets one or more areas to itself, depending on how many places it is divided in.
	state=0;  //0=in a wall, other=in an open area
	areacount=0;
	for(y=1; y<MAPSIZE_Y-1; y++)
	{
		for(x=1; x<MAPSIZE_X-1; x++)
		{
			if(state) {
				if( w->t[y][x].type==0 ) {
					state=0;
				} else {
					w->t[y][x].type = areacount;
				}
			} else {
				if( w->t[y][x].type!=0 ) {
					state=1, areacount++;
					w->t[y][x].type = areacount;
				}
			}
		}
		state=0;
	}

	// Merge overlapping areas
	// Apply room-connection algorithm
	connectareas( areacount );
	
	// Convert into tiles
	for(y=0; y<MAPSIZE_Y; y++)
		for(x=0; x<MAPSIZE_X; x++)
		{
			if( w->t[y][x].type )
				w->t[y][x].type = TILE_FLOOR;
			else
				w->t[y][x].type = TILE_WALL;
			
			w->t[y][x].flags = 0;
		}
}
//}}}
//{{{
void replace(int old, int n) {
	int i, ii;
	
	for(i=0; i<MAPSIZE_Y; i++)
		for(ii=0; ii<MAPSIZE_X; ii++)
			if( w->t[i][ii].type==old ) w->t[i][ii].type = n;
}
//}}}
//{{{
int connectareas(int num_areas)
{
	int tries=0, y, x, type, otype;
	
	// Find areas labelled unconnected but connected
	for(y=1; y<MAPSIZE_Y-1; y++)
		for(x=1; x<MAPSIZE_X-1; x++) {
			type = w->t[y][x].type;
			otype = w->t[y+1][x].type;
			if( type!=0 && otype!=0 && type!=otype ) {
				num_areas--;
				replace(otype, type);
			}
			otype = w->t[y][x+1].type;
			if( type!=0 && otype!=0 && type!=otype ) {
				num_areas--;
				replace(otype, type);
			}
		}
	
	// Build connections between areas
	do {
		// If a randomly-selected ROW connects two areas, make a tunnel
		y = RANGE(MAPSIZE_Y-2, 1);
		type = otype = 0;
		for(x=1; x<MAPSIZE_X-1; x++) {
			if( w->t[y][x].type ) {
				if(!otype)
					otype = w->t[y][x].type;
				else {
					type = w->t[y][x].type;
					if(type==otype) continue;
					
					for(;;x--) {
						if(w->t[y][x].type == otype) break;
						w->t[y][x].type = type;
					}
					replace(type, otype);
					num_areas--;
					break;
				}
			}
		}
		
		// If a randomly-selected COLUMN connects two areas, make a tunnel
		x = RANGE( MAPSIZE_X-2, 1);
		type = otype = 0;
		for(y=1; y<MAPSIZE_Y-1; y++) {
			if( w->t[y][x].type ) {
				if(!otype)
					otype = w->t[y][x].type;
				else {
					type = w->t[y][x].type;
					if(type==otype) continue;
					
					for(;;y--) {
						if(w->t[y][x].type == otype) break;
						w->t[y][x].type = type;
					}
					replace(type, otype);
					num_areas--;
					break;
				}
			}
		}
		
		tries++;
		if(tries>750) return 0;
		
	} while(num_areas>1);
	
	return 1;
}
//}}}
//{{{
void fill_room(room r)
{
	int which_room;
	int x, y;
	int width, height;
	
	width = r.right - r.left + 1;
	height = r.bottom - r.top + 1;
	
retry:
	which_room = RANGE(16,0);
	switch(which_room)
	{
		case 0: // Checkerboard
			// ##########
			// ## # # # #
			// # # # # ##
			// ## # # # #
			// # # # # ##
			// ## # # # #
			// ##########
			for(y=r.top; y<=r.bottom; y++)
				for(x=r.left+(y%2); x<=r.right; x+=2)
					w->t[y][x].type = TILE_WALL;
			break;
			
		case 1: // Pillars
		case 2:
		case 3:
		case 4:
			// #########
			// #       #
			// # # # # #
			// #       #
			// # # # # #
			// #       #
			// #########
			// Only for odd-width rooms with size >4
			if( !(width%2) || height<=4 )
				goto retry;
			
			y=r.top + 1;
			for(x=r.left+1; x<r.right; x+=2)
				w->t[y][x].type = TILE_WALL;
			
			if(r.bottom-r.top > 3)
			{
				y = r.bottom - 1;
				for(x=r.left+1; x<r.right; x+=2)
					w->t[y][x].type = TILE_WALL;
			}
			break;
			
		case 5: // Smoke-filled room
			// ##########
			// #~~~~~~~~#
			// #~~~~~~~~#
			// #~~~~~~~~#
			// #~~~~~~~~#
			// #~~~~~~~~#
			// ##########
			for(y=r.top; y<=r.bottom; y++)
				for(x=r.left; x<=r.right; x++)
					w->t[y][x].type = TILE_SMOKE;
			break;	
		case 6: // Trap room
			// ##########
			// #^^^^^^^^#
			// #^^^^^^^^#
			// #^^^^^^^^#
			// #^^^^^^^^#
			// #^^^^^^^^#
			// ##########
			if(w->level < 5) // No trap rooms above level 5
				goto retry;

			for(y=r.top; y<=r.bottom; y++)
				for(x=r.left; x<=r.right; x++)
					place_trap(x, y);
			break;

	}
	
	// 50% chance: intrinsically light room
	if(RANGE(1,0)==0)
	{
		for(y=r.top-1; y<=r.bottom+1; y++)
			for(x=r.left-1; x<=r.right+1; x++)
				w->t[y][x].flags |= TFLAG_INTRINSIC_LIGHT;
		
		illuminate( r.left + width/2, r.top + height/2, 400);
	}
}
//}}}
