#include "crogue.h"


int monstcanmove(int monst, int x, int y);
void monstmove(sint i);
void monstmoveto(int monst, int x, int y);

void monstmoverandomly(sint i, sint *xtarget, sint *ytarget);
void monstmovetowardsplayer(sint i, sint *xtarget, sint *ytarget);

/*
 *  Creature AI rules (simple)
 *
 * If can't sense player, do nothing
 * If can attack player, do so
 * Move towards player (Robots style)
 *
 *  Creature AI rules (better)
 *
 * If health<10% and speed>=player speed
 *   Switch AI to cowardly
 * If can attack player, do so
 * If can sense player, move towards player (Robots style)
 * Else if can move in same direction as on previous turn
 *   Move in same direction as previous turn
 * Else
 *   Move randomly
 *
 *  Creature AI rules (cowardly)
 *
 * Consider all adjacent tiles; choose whichever is farthest from the player
 * If new tile's distance from plr = 1, switch to 'aggressive' step 2
 *
 *  Creature AI rules (stalk)
 *
 * Consider all adjacent tiles; choose whichever unlit tile is closest
 * If no unlit adjacent tiles, switch rules to aggressive and signal all stalkers to attack
 */

void monst_attack_type(int trigger, int monst, int always_hit);

// Utility
//{{{
int monstcanmove(int monst, int x, int y)
{
	if(x==w->m[monst].x && y==w->m[monst].y) // Standing still is always legal
		return 1;
	
	return (
	         (w->tiledescs[w->t[y][x].type].passable) &&  // Only passable tiles
	         !(w->t[y][x].flags & TFLAG_OCCUPIED)         // Only empty tiles
	       ) ||
	       (x==w->plr.x && y==w->plr.y);                 // Always allow attacking the player
}
//}}}
//{{{
#define NUM_HALLUNAMES 6
const char* hallunames[NUM_HALLUNAMES] = {
	"bugblatter beast of Traal",
	"greater hell beast",
	"invisible asmodeus",
	"power grid bug",
	"pet kitten",
	"Usenet troll"
};
//}}}
//{{{
const char* monstname(int n)
{
	if(!w->plr.hallutime)
		return CONSTFILEDEREF(MDESC(n)->name);
	else {
		return hallunames[RANGE(NUM_HALLUNAMES-1, 0)];
	}
}
//}}}
//{{{
int monstbytile(ushort x, ushort y)
{
	int i;
	
	if(!w->t[y][x].flags & TFLAG_OCCUPIED)
		return -1;
	
	for(i=0; i<MONSTERS_MAX; i++)
	{
		if(!isNull(w->m[i].type) && w->m[i].x==x && w->m[i].y==y)
			return i;
	}
	return -1;
}
//}}}

//{{{
void compress_monsters(void)
{
	int i, j;
	
	for(i=j=0; i<MONSTERS_MAX; i++)
	{
		if(!isNull(w->m[i].type))
		{
			memcpy(&w->m[j], &w->m[i], sizeof(monster));
			j++;
		}
	}
}
//}}}

//{{{
int monst_hit_plr(int attacknum, int monster)
{
	if(15 + MATTACK(monster, attacknum).tohit + MDESC(monster)->level - (sshort)RANGE(20,1) >= w->plr.extrinsic[STAT_DV])
		return 1;
	else
	{
		message("The %s misses!", monstname(monster));
		return 0;
	}
}
//}}}
//{{{
void monst_attack_plr(int attacknum, int monster)
{
	int whichitem;
	int i;
	monstattack *attacks;
	
	attacks = MDESC(monster)->attacks;
	
	message("The %s", monstname(monster));
	
	switch(attacks[attacknum].type)
	{
		case MATTACK_NORM:
attack_normal:
			message("hits!");
			break;
		case MATTACK_ZAP:
			message("zaps you!");
			break;
		case MATTACK_BITE:
			message("bites!");
			break;
		case MATTACK_CLAW:
			message("claws you!");
			break;
		case MATTACK_KICK:
			message("kicks!");
			break;
		case MATTACK_POIS_BITE:
			message("bites!");
			goto attack_poison;
		case MATTACK_POIS:
			message("poisons you!");
attack_poison:
			if(RANGE(20, 1)>w->plr.extrinsic[STAT_TOUGHNESS])
				plr_mod_stat(0, -RANGE(2,1));
			break;
		case MATTACK_PARAL:
			message("paralyzes you!");
			if(!w->plr.paraltime)
				w->plr.paraltime = rrandom(attacks[attacknum].damage);
			break;
		case MATTACK_PARALONLY:
			message("paralyzes you!");
			if(!w->plr.paraltime)
				w->plr.paraltime = rrandom(attacks[attacknum].damage);
			goto skip_damage;
		case MATTACK_STEALITEM:
			if(!w->plr.numitems) // If no items around to steal, convert to a normal attack.
				goto attack_normal;
			
			whichitem = rand()%w->plr.numitems;
			message("steals your %s. The %s disappears!", itemname(&INVENTORY(whichitem)), monstname(monster));
			
			INVENTORY(whichitem).type = 0;
			w->plr.numitems--;
			update_player();
			
			monst_remove(monster);
			break;
		case MATTACK_RUST:
			message("hits!");
			
			for(i=0; i<PLRINVSIZE; i++)
			{
				if(!INVENTORY(i).type || !(INVENTORY(i).flags & ITEMFLAG_EQUIPPED))
					continue;
				if(ITEMDESC(INVENTORY(i)).flags & ITEMFLAG_RUSTPROOF)
					continue;
				if(RANGE(3,1)!=1)
					continue;
				
				// FIXME: Handle weapon type
				message("Your %s rusts!", itemname(&INVENTORY(i)));
				INVENTORY(i).info.armr.plus--;
			}
			break;
		case MATTACK_POISSTING:
			message("stings!");
			goto attack_poison;
		case MATTACK_STEALGOLD:
		default:
			message("hits (invalid attack!).");
			break;
	}
	plr_takedamage(rrandom( attacks[attacknum].damage ));
skip_damage:
	;
}
//}}}

//{{{
void monst_attack_type(int trigger, int monst, int always_hit)
{
	int i;
	
	for(i=0; i<MDESC(monst)->numattacks; i++)
	{
		if(MATTACK(monst, i).trigger == trigger)
			if(always_hit || monst_hit_plr(i, monst))
				monst_attack_plr(i, monst);
		
		if(isNull(w->m[monst].type))
			return;
	}
}
//}}}

//{{{
void monstmoveto(int monst, int x, int y)
{
	
	if(x==w->plr.x && y==w->plr.y)
	{
		monst_attack_type(MTRIGGER_NORMAL, monst, 0);
	} else {
		w->t[w->m[monst].y][w->m[monst].x].flags &= ~TFLAG_OCCUPIED;
		draw_tile(w->m[monst].x, w->m[monst].y);
		w->m[monst].y = y;
		w->m[monst].x = x;
		w->t[y][x].flags |= TFLAG_OCCUPIED;
		draw_tile(x, y);
	}
}
//}}}
//{{{
void monstmove(sint i)
{
	sint best_x, best_y;
	
	if( MDESC(i)->flags & MONSTFLAG_REGENERATE && w->m[i].hps < MDESC(i)->hps_max)
	{
		w->m[i].hps += 2;
		
		if(w->t[w->m[i].y][w->m[i].x].flags & TFLAG_LIT)
			message("The %s regenerates!", monstname(i));
	}
	
	if( MDESC(i)->flags & MONSTFLAG_MOVERANDOMLY )	// Some monsters don't have AI; they just move about
	{
		monstmoverandomly(i, &best_x, &best_y);
	}
	else if( distancesquare(w->m[i].x, w->m[i].y, w->plr.x, w->plr.y) <= 144  && // Reasonably near the player (12 tiles)
	        (tracevision(w->plr.x, w->plr.y, w->m[i].x, w->m[i].y) ||            // And can see the player
	         distancesquare(w->m[i].x, w->m[i].y, w->plr.x, w->plr.y) <= 16) )   // OR can hear the player (dist<=4)
	{
		monstmovetowardsplayer(i, &best_x, &best_y);
	}
	else
	{
		monstmoverandomly(i, &best_x, &best_y);
	}

	monstmoveto(i, best_x, best_y);
}
//}}}
//{{{
void monstmoverandomly(sint i, sint *xtarget, sint *ytarget)
{
	do
	{
		*xtarget = w->m[i].x + RANGE(2, 0) - 1;
		*ytarget = w->m[i].y + RANGE(2, 0) - 1;
	} while( !monstcanmove(i, *xtarget, *ytarget) );
}
//}}}
//{{{
void monstmovetowardsplayer(sint i, sint *xtarget, sint *ytarget)
{
	sint best_val = 1000;
	sint best_x = w->m[i].x, best_y = w->m[i].y;
	sint xi, yi;

	for(yi = w->m[i].y-1; yi <= w->m[i].y+1; yi++)
		for(xi = w->m[i].x-1; xi <= w->m[i].x+1; xi++)
		{
			if( monstcanmove( i, xi, yi ) &&  // Move is legal
				distancesquare(xi, yi, w->plr.x, w->plr.y) < best_val ) // And is an improvement
			{
				best_val = distancesquare(xi, yi, w->plr.x, w->plr.y);
				best_x = xi;
				best_y = yi;
			}
		}
	
	*xtarget = best_x;
	*ytarget = best_y;
}
//}}}

//{{{
void addrandommonst(void)
{
	int x, y;
	
	// Find an open tile
	while(1)
	{
		x = RANGE( MAPSIZE_X-1, 1);
		y = RANGE( MAPSIZE_Y-1, 1);
		if( w->t[y][x].type==TILE_FLOOR && !(w->t[y][x].flags & (TFLAG_OCCUPIED)) )
		{
			addmonster(x, y);
			draw_tile(x, y);
			return;
		}
	}
}
//}}}
//{{{
void addmonster(ushort x, ushort y)
{
	int i;
	
	// Find an open monster handle
	for(i=0; i<MONSTERS_MAX; i++) {
		if(isNull(w->m[i].type))
			break;
	}
	if(i>=MONSTERS_MAX)
		return;
	
	w->m[i].type = deref_file_ptr_partial(w->mapdescs[w->level].monstertab);
	w->m[i].flags = 0;
	w->m[i].hps = MDESC(i)->hps_max;
	w->m[i].x = x;
	w->m[i].y = y;
	w->t[y][x].flags |= TFLAG_OCCUPIED;
	
}
//}}}
//{{{
void monst_remove(int monstid)
{
	w->t[w->m[monstid].y][w->m[monstid].x].flags &= ~TFLAG_OCCUPIED;
	w->m[monstid].type = filelink_null;
	draw_tile(w->m[monstid].x, w->m[monstid].y);
}
//}}}
//{{{
int monst_takedamage(int monstid, int damage, int is_melee)
{
	if(damage<0)
		return 0;
	
	damage = max( damage-MDESC(monstid)->pv, damage/4 );
	w->m[monstid].hps -= damage;
	
	if( w->m[monstid].hps <= 0 )
	{
		message("You kill it!");
		awardXP(MDESC(monstid)->xp);
		
		if(is_melee)
		{
			// Look for death-defense attacks
			monst_attack_type(MTRIGGER_DEATH, monstid, 0);
		}
		
		monst_remove(monstid);
		return 1;
	} else {
		if(is_melee)
		{
			// Look for defensive attacks
			monst_attack_type(MTRIGGER_DEFENSE, monstid, 0);
		}
		return 0;
	}
}
//}}}
//{{{
void monst_heal(int monstid, int amt)
{
	w->m[monstid].hps += amt;
	
	if(w->m[monstid].hps > MDESC(monstid)->hps_max)
		w->m[monstid].hps = MDESC(monstid)->hps_max;
}
//}}}

