#include "crogue.h"

const unsigned char spellMenuTabStops[3] = { 0, 60, 85 };

static int numSpellsKnown(void);
static int spellByIndex(int n);


//{{{
void read_spellbook(int i)
{
	int difficulty, skill;
	int which_spell = ITEMDESC(INVENTORY(i)).desc.spellbook.spellid;
	
	message("You read the %s.", itemname(&INVENTORY(i)));
	
	timepass(w->spelldescs[which_spell].pp_cost);
	
	difficulty = w->spelldescs[which_spell].difficulty;
	skill = w->playerclasses[w->plr.class].spellcasting/3 + w->plr.level + RANGE(5,1);
	
	if(difficulty>skill)
	{
		message("You can't understand the contents of this spellbook.");
		return;
	}
	
	identify_type(INVENTORY(i).type);
	
	message("You learn the %s spell.", (const char*)CONSTFILEDEREF(w->spelldescs[which_spell].name) );
	
	w->plr.spellknowledge[which_spell] += 3 * w->playerclasses[w->plr.class].spellcasting;
	
	INVENTORY(i).info.book.readings--;
	if(INVENTORY(i).info.book.readings<=0)
	{
		message("The spellbook suddenly disappears.");
		INVENTORY(i).type = 0;
		w->plr.numitems--;
	}
}
//}}}
//{{{
void init_spells(void)
{
	int i;
	
	w->plr.spellknowledge = (int*)malloc( sizeof(int)*w->desc.numspells );
	if(!w->plr.spellknowledge)
		fatal("Out of memory.");
	
	for(i=0; i<w->desc.numspells; i++)
	{
		w->plr.spellknowledge[i] = 0;
	}
}
//}}}
//{{{
static int numSpellsKnown(void)
{
	int i, n=0;
	
	for(i=0; i<w->desc.numspells; i++)
		if(w->plr.spellknowledge[i]) n++;
	
	return n;
}
//}}}
//{{{
static int spellByIndex(int n)
{
	int i;
	
	for(i=0; n>=0; i++) {
		if(w->plr.spellknowledge[i] > 0)
			n--;
	}
	
	return i-1;
}
//}}}
//{{{
void spellString(int n, char *buf)
{
	spelldesc *s;
	
	n = spellByIndex(n);
	s = &w->spelldescs[n];
	
	sprintf(buf, "%s\t"   "%i pp\t"  "%i", (const char*)CONSTFILEDEREF(s->name), s->pp_cost, w->plr.spellknowledge[n]);
}
//}}}
//{{{
void cast_spell(void)
{
	int which;
	
	if(!numSpellsKnown())
	{
		message("You don't know any spells.");
		return;
	}
	
	setTabStops(3, spellMenuTabStops);
	which = UI_Menu_Pick( COORD(7,14), COORD(142,79), numSpellsKnown(), &spellString, 0 );
	UI_TF_FullRedraw(&w->tilefield);
	
	if(which==-1)
	{
		message("Never mind.");
		return;
	}
	
	which = spellByIndex(which);
	
	if(w->plr.pps < w->spelldescs[which].pp_cost)
	{
		message("You don't have enough power to cast that spell.");
		return;
	}
	
	w->plr.spellknowledge[which]--;
	w->plr.pps -= w->spelldescs[which].pp_cost;
	
	switch(w->spelldescs[which].effect)
	{
		case SPELL_BOLT:
			bolt( w->plr.x, w->plr.y, prompt_dir("In what direction"), 6 );
			break;
		case SPELL_LIGHT:
			message("A lit field surrounds you.");
			illuminate(w->plr.x, w->plr.y, 36);
			calc_light();
			break;
		case SPELL_IDENTIFY:
			identify();
			break;
		case SPELL_TELEPORT:
			teleport_player();
			break;
	}
	
	timepass(1);
}
//}}}
