#include "crogue.h"

void GraySprite8_BLIT(short x,short y,short h, const unsigned char* sprite1, const unsigned char* sprite2, unsigned char maskval, void* dest1, void* dest2);

void _UI_TF_DrawTile(UI_Tilefield *tf, int x, int y, UI_TF_TileInfo tile);
void _UI_TF_DrawSprite(UI_Tilefield *tf, int x, int y, void *plane, const char *graphics, short mode);

//{{{
void UI_TF_MoveCamera(UI_Tilefield *tf, int x, int y)
{
	int oldx=tf->viewoffset.x, oldy=tf->viewoffset.y;
		// Previous x/y offset, to compare later to see if it's changed
	
	// Scroll X
	if( x-6 <= tf->viewoffset.x || x+6-tf->viewoffset.x >= tf->window.x )
	{
		tf->viewoffset.x = max(x - ((sint)tf->window.x)/2, 0);
		if(tf->viewoffset.x + tf->window.x > tf->size.x)
			tf->viewoffset.x = tf->size.x - tf->window.x;
		
	}
	
	// Scroll Y
	if( y-4 <= tf->viewoffset.y || y+4-tf->viewoffset.y >= tf->window.y )
	{
		tf->viewoffset.y = max(y - ((sint)tf->window.y)/2, 0);
		if(tf->viewoffset.y + tf->window.y > tf->size.y)
			tf->viewoffset.y = tf->size.y - tf->window.y;
	}
	
	if( tf->viewoffset.x != oldx || tf->viewoffset.y != oldy ) UI_TF_FullRedraw(tf);
}
//}}}
//{{{
void UI_TF_PlaceCamera(UI_Tilefield *tf, int x, int y)
{
	tf->viewoffset.x = max(x - ((sint)tf->window.x)/2, 0);
	tf->viewoffset.y = max(y - ((sint)tf->window.y)/2, 0);
	if(tf->viewoffset.x + tf->window.x > tf->size.x)
		tf->viewoffset.x = tf->size.x - tf->window.x;
	if(tf->viewoffset.y + tf->window.y > tf->size.y)
		tf->viewoffset.y = tf->size.y - tf->window.y;
	return;
}
//}}}
//{{{
void UI_TF_Scroll(UI_Tilefield *tf, int x, int y)
{
	signed int new_x, new_y;
	
	new_x = tf->viewoffset.x + x;
	new_y = tf->viewoffset.y + y;
	if(new_x<0) new_x = 0;
	if(new_y<0) new_y = 0;
	if(new_x + tf->window.x > tf->size.x) new_x = tf->size.x - tf->window.x;
	if(new_y + tf->window.y > tf->size.y) new_y = tf->size.y - tf->window.y;
	
	tf->viewoffset.x = new_x;
	tf->viewoffset.y = new_y;
	
	UI_TF_FullRedraw(tf);
}
//}}}
//{{{
int UI_TF_Init(UI_Tilefield *tf, coord pos, coord size, coord window, const uchar *graphics, uchar tileheight, uchar tilewidth)
{
	int i, j;
	
	tf->pos = pos;
	tf->size = size;
	tf->window = window;
	tf->graphics = graphics;
	tf->tileheight = tileheight;
	tf->tilewidth = tilewidth;
	tf->viewoffset = COORD(0,0);
	
	// In case of failure, make sure it's obvious when cleaning up that it's not alloced
	// The tf->tiles doesn't need this because, being first, it's guaranteed to be initialized
	// with something, even if it's null.
	if( !(tf->tiles = (UI_TF_TileInfo**)malloc( sizeof(UI_TF_TileInfo*) * tf->size.y )) )
		return 0;	// Malloc failed!
	
	for(i=0; i<tf->size.y; i++) {
		tf->tiles[i] = (UI_TF_TileInfo*)malloc(sizeof(UI_TF_TileInfo) * tf->size.x);
		if( !tf->tiles[i] )
			goto cleanup;
		for(j=0; j<tf->size.x; j++) {
			tf->tiles[i][j].lighting = LIGHTING_UNINITIALIZED;
		}
	}
	
	return 1;

cleanup:
	UI_TF_Deinit(tf);
	return 0;
}
//}}}
//{{{
// FIXME: this function will crash if called before UI_TF_Init
// (ie, when fatal() is called during initialization)
void UI_TF_Deinit(UI_Tilefield *tf)
{
	int i;
	
	if( tf->tiles ) {
		for(i=0; i<tf->size.y; i++) {
			if( tf->tiles[i] != NULL ) {
				free(tf->tiles[i]);
			} else {
				break;
			}
		}
		free(tf->tiles);
		tf->tiles = 0;
	}
}
//}}}
//{{{
void UI_TF_PutTile(UI_Tilefield *tf, int x, int y, uchar tnum, uchar lighting)
{
	int tx, ty;
	tx = x - tf->viewoffset.x;
	ty = y - tf->viewoffset.y;
	
	if( tf->tiles[y][x].tile != tnum || tf->tiles[y][x].lighting != lighting ) {
		tf->tiles[y][x].tile = tnum;
		tf->tiles[y][x].lighting = lighting;
		
		if(tx<0 || ty<0 || tx>=tf->window.x || ty>=tf->window.y) return;
		
#ifdef IS_CALCULATOR
		_UI_TF_DrawTile(tf, tx, ty, tf->tiles[y][x]);
#else
		gotoxy(x, y+2);
		addch(tiles[tnum]);
#endif
	}
}
//}}}
//{{{
void UI_TF_FullRedraw(UI_Tilefield *tf)
{
	int i, ii;

#ifdef IS_CALCULATOR	
	for(i=0; i<tf->window.y; i++)
		for(ii=0; ii<tf->window.x; ii++)
			_UI_TF_DrawTile(tf, ii, i, tf->tiles[i+tf->viewoffset.y][ii+tf->viewoffset.x]);
#else
	for(i=0; i<MAPSIZE_Y; i++)
	{
		for(ii=0; ii<MAPSIZE_X; ii++)
		{
			gotoxy(ii, i+2);
			addch(tiles[tf->tiles[i][ii].tile]);
		}
	}
#endif
}
//}}}
//{{{
void _UI_TF_DrawTile(UI_Tilefield *tf, int x, int y, UI_TF_TileInfo tile)
{
#ifdef IS_CALCULATOR
	void *light = GrayGetPlane(LIGHT_PLANE);
	void *dark = GrayGetPlane(DARK_PLANE);
	const uchar *lighttile;
	const uchar *darktile;
	
	switch(tile.lighting)
	{
		case LIGHTING_LIT:
			lighttile = darktile = tf->graphics + tile.tile*tf->tileheight;
			break;
		case LIGHTING_DARK:
			lighttile = blacktile;
			darktile = tf->graphics + tile.tile*tf->tileheight;
			break;
		case LIGHTING_UNEXPLORED:
			lighttile = whitetile;
			darktile = blacktile;
			break;
		default:
			return;
	}
	GraySprite8_BLIT(tf->pos.x + x * tf->tilewidth,
		tf->pos.y + y * tf->tileheight,
		tf->tileheight,
		lighttile,
		darktile,
		0b00000111,
		light,
		dark);

/*	switch(tile.lighting)
	{
		case LIGHTING_LIT:
			_UI_TF_DrawSprite(tf, x, y, dark, tilemask, SPRT_AND);
			_UI_TF_DrawSprite(tf, x, y, dark, (char*)(tf->graphics + tile.tile*tf->tileheight), SPRT_OR);
			_UI_TF_DrawSprite(tf, x, y, light, tilemask, SPRT_AND);
			_UI_TF_DrawSprite(tf, x, y, light, (char*)(tf->graphics + tile.tile*tf->tileheight), SPRT_OR);
			break;
		case LIGHTING_DARK:
			_UI_TF_DrawSprite(tf, x, y, dark, tilemask, SPRT_AND);
			_UI_TF_DrawSprite(tf, x, y, dark, (char*)(tf->graphics + tile.tile*tf->tileheight), SPRT_OR);
			_UI_TF_DrawSprite(tf, x, y, light, blacktile, SPRT_OR);
			break;
		case LIGHTING_UNEXPLORED:
			_UI_TF_DrawSprite(tf, x, y, dark, blacktile, SPRT_OR);
			_UI_TF_DrawSprite(tf, x, y, light, tilemask, SPRT_AND);
			break;
	}*/
#else
	gotoxy(x, y);
	addch(tile.tile);
#endif
}
//}}}

//{{{
void GraySprite8_BLIT(short x,short y,short h,
                      const unsigned char* sprite1,
                      const unsigned char* sprite2,
                      unsigned char maskval,
                      void* dest1,
                      void* dest2)
{
	register long           offset = (y<<5)-(y<<1)+((x>>3)&0x1e);
	register long           addr1  = (long)dest1+offset;
	register long           addr2  = (long)dest2+offset;
	register unsigned short cnt      = 24-(x&15);
	register long           maskval2 = ~((long)(~(maskval)&0xff)<<cnt);
		for (;h;h--,addr1+=30,addr2+=30) {
		*(long*)addr1&=maskval2;
		*(long*)addr1|=(long)(*sprite1++)<<cnt;
		*(long*)addr2&=maskval2;
		*(long*)addr2|=(long)(*sprite2++)<<cnt;
	}
}
//}}}

//{{{
/*void _UI_TF_DrawSprite(UI_Tilefield *tf, int x, int y, void *plane, const char *graphics, short mode)
{
#ifdef IS_CALCULATOR
	Sprite8(
		tf->pos.x + x * tf->tilewidth,
		tf->pos.y + y * tf->tileheight,
		tf->tileheight,
		(char*)graphics,
		plane,
		mode );
#endif
}*/
//}}}
